(* This program is free software; you can redistribute it and/or modify it under  *)
(* the terms of the GNU General Public License as published by the Free Software  *)
(* Foundation; either version 2 of the License, or (at your option) any later     *)
(* version.                                                                       *)
(*                                                                                *)  
(* This program is distributed in the hope that it will be useful, but WITHOUT    *)
(* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  *)
(* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. *)
(*                                                                                *)
(* Programmed by Alok Menghrajani.                                                *)
(* http://www.alokonline.com/projets/compiler/simpleb.html                        *)
(*                                                                                *)
(* Description: this file contains the function translate, which takes an abs syn *)
(*              tree (generated by Grammar.program) and translates it into a seq  *)
(*              of intermediate representation trees (Tree.stm list). Ordinary,   *)
(*              translate would return only one Treee.stm, which would then be    *)
(*              broken down into a list of trees by Canon.linerize. However,      *)
(*              I break into a list here. Also, unlike Appel who uses the         *)
(*              constructors Ex, Nx and Cx to assist in translating expressions,  *)
(*              statements and conditionals, we break our translate function down *)
(*              into translate_exp, translate_assign and translate_list. We also  *)
(*              keep a symbol table that associates named variables with          *)
(*              temporary variables, which will later be used as stack slots.     *)
(*                                                                                *)
(*              Note: if a named variable is used on the right side of an assign  *)
(*              expression before it is declared and inserted into the symbol     *)
(*              table, Errormsg.error is called.                                  *)
(*                                                                                *)
(* X-reference: tree.ml, absyn.ml,  errormsg.ml, symbol.ml                        *)


module T = Tree
module A = Absyn
module S = Symbol

let tbl = S.empty ()

let errorLabel = Label.newlabel()

(* let _ = S.enter tbl (S.symbol "output") (T.TEMP (Temp.newtemp())) *)

exception Error of string

type exp = Ex of Tree.exp
         | Nx of Tree.stm
         | Cx of (Label.label * Label.label -> Tree.stm)

let unEx e =
     match e with 
       Ex e -> e
     | Nx s -> T.ESEQ(s,T.CONST Int32.zero,T.PRE)
     | Cx c ->
        let r = Temp.newtemp() in
        let t = Label.newlabel() and f = Label.newlabel() in
            T.ESEQ(T.SEQ(T.MOVE(T.TEMP r, T.CONST Int32.one),
		     T.SEQ(c(t,f),
		       T.SEQ(T.LABEL f,
			 T.SEQ(T.MOVE(T.TEMP r, T.CONST Int32.zero),T.LABEL t)))),
		   T.TEMP r, T.PRE);;

let unNx s =
     match s with
       Ex e -> T.EXP(e)
     | Nx s -> s
     | Cx c -> raise (Error "UnNx of Cx");;

let unCx c =
     match c with
       Ex e -> let f (l1, l2) = T.CJUMP(T.EQ, e, T.CONST Int32.zero, l2, l1) in
               f
     | Nx s -> raise (Error "UnCx of Nx")
     | Cx c -> c;;

let check_var var =
      match var with
        A.SimpleVar(v, pos) ->
          (match S.look tbl v with
             Some(binding) -> binding
           | None -> Errormsg.error pos ("Variable " ^ (S.name v) ^ " is undefinied");
                     raise (Error "Compilation exited due to errors."))
;;


let rec translate_boolexp (b, break, continue) =
      let translate_cmpop op =
            (match op with
              A.EqOp -> T.EQ
            | A.NeOp -> T.NE
            | A.LtOp -> T.LT
            | A.LeOp -> T.LE
            | A.GtOp -> T.GT
            | A.GeOp -> T.GE) in
      match b with
        A.True _ ->  let f(t, f)=T.JUMP(t) in
                     Cx(f)
      | A.False _ -> let f(t, f)=T.JUMP(f) in
                     Cx(f)
      | A.Unopb (be, uop, _) -> (match uop with
                                  A.NotOp -> let f1=unCx(translate_boolexp(be, break, continue)) in
                                             let f2(t, f)=f1(f, t) in
                                             Cx(f2))
      | A.Binopb (left, bop, right, _) -> let f1=unCx(translate_boolexp(left, break, continue)) and
                                              f2=unCx(translate_boolexp(right, break, continue)) and
                                              f2_hack=unCx(translate_boolexp(right, break, continue)) in
(* HACK used to avoid creation of identical labels *)
                                          (match bop with
                                            A.AndOp -> let z = Label.newlabel() in
                                                       let f(t, f)=T.SEQ(f1(z, f),
                                                                    T.SEQ(T.LABEL z, f2(t, f))) in
                                                       Cx(f)
                                          | A.XorOp -> let z1=Label.newlabel() in
                                                       let z2=Label.newlabel() in
                                                       let f(t, f)=T.SEQ(f1(z1, z2),
                                                                     T.SEQ(T.LABEL z1,
                                                                       T.SEQ(f2(f, t),
                                                                         T.SEQ(T.LABEL z2, f2_hack(t, f))))) in
                                                      Cx(f)
                                          | A.OrOp -> let z=Label.newlabel() in
                                                      let f(t, f)= T.SEQ(f1(t, z),
                                                                     T.SEQ(T.LABEL z, f2(t, f))) in
                                                      Cx(f))
      | A.Cmpop (left, op, right, _) ->  let f(t, f)=T.CJUMP(translate_cmpop op, unEx(translate_exp(left, break, continue)),
                                                             unEx(translate_exp(right, break, continue)), t, f) in
                                         Cx(f)



and translate_exp (e, break, continue) =
      let translate_op op = 
        match op with
          A.PlusOp -> T.PLUS
        | A.MinusOp -> T.MINUS
        | A.TimesOp -> T.MUL
        | A.DivideOp -> T.DIV
        | A.ModOp -> T.MOD in
       match e with
         A.IntExp c -> Ex(T.CONST c)
       | A.VarExp v -> let binding = check_var v in
                       Ex(T.TEMP binding) 
       | A.OpExp (left, op, right, _) -> let l1 = unEx(translate_exp(left, break, continue)) and
                                             l2 = unEx(translate_exp(right, break, continue)) in
                                         let o = translate_op op in
                                         Ex(T.BINOP(o, l1, l2))
       | A.PreIncExp (v, _) -> let binding = check_var v in
                               Ex(T.ESEQ(
                                    T.MOVE(T.TEMP binding, T.BINOP(T.PLUS, T.TEMP binding, T.CONST Int32.one)),
                                    T.TEMP binding, T.PRE))
       | A.PostIncExp (v, _) -> let binding = check_var v in
(* HACK FOR TUE DEADLINE         Ex(T.ESEQ(
                                    T.MOVE(T.TEMP binding, T.BINOP(T.PLUS, T.TEMP binding, T.CONST Int32.one)),
                                    T.TEMP binding, T.POST))  *)
                                let t=Temp.newtemp() in
                                 Ex(T.ESEQ(T.SEQ(T.MOVE(T.TEMP t, T.TEMP binding),
                                             T.MOVE(T.TEMP binding, T.BINOP(T.PLUS, T.TEMP binding, T.CONST Int32.one))),
                                             T.TEMP t, T.PRE))

       | A.PreDecExp (v, _) -> let binding = check_var v in
                               Ex(T.ESEQ(
                                    T.MOVE(T.TEMP binding, T.BINOP(T.MINUS, T.TEMP binding, T.CONST Int32.one)),
                                    T.TEMP binding, T.PRE))

       | A.PostDecExp (v, _) -> let binding = check_var v in
(* HACK FOR TUE DEADLINE         Ex(T.ESEQ(
                                    T.MOVE(T.TEMP binding, T.BINOP(T.MINUS, T.TEMP binding, T.CONST Int32.one)),
                                    T.TEMP binding, T.POST))  *)
                                let t=Temp.newtemp() in
                                 Ex(T.ESEQ(T.SEQ(T.MOVE(T.TEMP t, T.TEMP binding),
                                             T.MOVE(T.TEMP binding, T.BINOP(T.MINUS, T.TEMP binding, T.CONST Int32.one))),
                                             T.TEMP t, T.PRE))

and translate_stmt (s, break, continue) =
      match s with
        A.SeqStmt l -> (match l with
                          (head, _)::tail -> let h = unNx(translate_stmt(head, break, continue)) in
                                             let t = unNx(translate_stmt(A.SeqStmt(tail), break, continue)) in
					     Nx(T.SEQ(h, t))
		       | [] -> Nx(T.NIL))

      | A.AssignStmt(v, e, _) -> let right_ir = unEx(translate_exp(e, break, continue)) in
                                                   let binding = check_var v in
						   Nx(T.MOVE(T.TEMP binding, right_ir))

      | A.IfStmt(be, t, f, _) -> let lt = Label.newlabel() in
                                 let lf = Label.newlabel() in
                                 let le = Label.newlabel() in
                                 let b = unCx(translate_boolexp(be, break, continue)) in
                                 Nx(T.SEQ(b(lt, lf),
                                      T.SEQ(T.LABEL lt,
                                        T.SEQ(unNx(translate_stmt(t, break, continue)),
                                          T.SEQ(T.JUMP le,
                                            T.SEQ(T.LABEL lf,
                                              T.SEQ(unNx(translate_stmt(f, break, continue)), T.LABEL le)))))))

      | A.WhileStmt(be, s, _) -> let loopstart = Label.newlabel() in
                                 let loopfirststmt = Label.newlabel() in
                                 let loopbreak = Label.newlabel() in
                                 let b = unCx(translate_boolexp(be, break, continue)) in
                                 Nx(T.SEQ(T.LABEL loopstart,
                                      T.SEQ(b(loopfirststmt, loopbreak),
                                        T.SEQ(T.LABEL loopfirststmt,
                                          T.SEQ(unNx(translate_stmt(s, loopbreak, loopstart)),
                                            T.SEQ(T.JUMP loopstart, T.LABEL loopbreak))))))

      | A.ForStmt(v, e1, be, it, st, _) -> let binding = check_var v in
                                           let loopstart = Label.newlabel() in
                                           let loopfirststmt = Label.newlabel() in
                                           let loopcontinue = Label.newlabel() in
                                           let loopbreak = Label.newlabel() in
                                           let right_ir = unEx(translate_exp(e1, break, continue)) in
                                           let b = unCx(translate_boolexp(be, break, continue)) in
                                           Nx(T.SEQ(T.MOVE(T.TEMP binding, right_ir),
                                                T.SEQ(T.LABEL loopstart,
                                                  T.SEQ(b(loopfirststmt, loopbreak),
                                                    T.SEQ(T.LABEL loopfirststmt,
                                                      T.SEQ(unNx(translate_stmt(st, loopbreak, loopcontinue)),
                                                        T.SEQ(T.LABEL loopcontinue,
                                                          T.SEQ(unNx(translate_stmt(it, loopbreak, loopcontinue)),
                                                            T.SEQ(T.JUMP loopstart, T.LABEL loopbreak)))))))))
      | A.BreakStmt(pos) -> if (break==errorLabel) then
                            (Errormsg.error pos ("Illegal break");
			    raise (Error "Compilation exited due to errors."))
                          else Nx(T.JUMP break)

      | A.ContinueStmt(pos) -> if (break==errorLabel) then
                               (Errormsg.error pos ("Illegal continue");
			       raise (Error "Compilation exited due to errors."))
                             else Nx(T.JUMP continue)

      | A.WriteintStmt(e, _) -> let right_ir = unEx(translate_exp(e, break, continue)) in
                                Nx(T.WRITEINT right_ir)

let rec translate_decl d =
      match d with
        (v, e, _)::tail -> let temp = Temp.newtemp() in
	                   let right_ir = unEx(translate_exp(e, errorLabel, errorLabel)) in
	                   S.enter tbl v (temp); 
			   Nx(T.SEQ(T.MOVE(T.TEMP temp, right_ir), unNx(translate_decl tail)))
      | [] -> Nx(T.NIL)
;;

let translate p =
     match p with
       A.Program(A.VarDecSeq dec, A.SeqStmt s, context) ->
	 Errormsg.setContext context;
	 (* handle variable declarations *)
	 let d = translate_decl(dec) in
(*         S.print_symbol_table tbl;  *)
         (* translate main code *)
         let main = translate_stmt(A.SeqStmt(s), errorLabel, errorLabel) in
         T.SEQ(unNx(d), unNx(main))
     | _ -> raise (Error("Invalid AST: SeqStmt not encountered"));;




